@@ -131,9 +131,13 @@ CK_RV do_EncryptDecryptRSA(struct GENERATED_TEST_SUITE_INFO *tsuite)
 
             if (tsuite->mech.mechanism == CKM_RSA_PKCS_OAEP &&
                  tsuite->tv[i].oaep_params.hashAlg != CKM_SHA_1 &&
-                 tsuite->tv[i].oaep_params.hashAlg != CKM_SHA256) {
+                 tsuite->tv[i].oaep_params.hashAlg != CKM_SHA224 &&
+                 tsuite->tv[i].oaep_params.hashAlg != CKM_SHA256 &&
+                 tsuite->tv[i].oaep_params.hashAlg != CKM_SHA384 &&
+                 tsuite->tv[i].oaep_params.hashAlg != CKM_SHA512) {
                  testcase_skip("CCA Token cannot use RSA OAEP with a hash "
-                              "algorithm other than SHA1 and SHA256");
+                              "algorithm other than SHA1 and SHA2: %s",
+                              mech_to_str(tsuite->tv[i].oaep_params.hashAlg));
                  free(s);
                  continue;
              }
@@ -180,6 +184,17 @@ CK_RV do_EncryptDecryptRSA(struct GENERATED_TEST_SUITE_INFO *tsuite)
             }
         }
 
+        if (tsuite->mech.mechanism == CKM_RSA_PKCS_OAEP) {
+            if (!mech_supported(slot_id, tsuite->tv[i].oaep_params.hashAlg)) {
+                testcase_skip("Slot %u doesn't support OAEP hash alg %s (0x%x)",
+                              (unsigned int)slot_id,
+                              mech_to_str(tsuite->tv[i].oaep_params.hashAlg),
+                              (unsigned int)tsuite->tv[i].oaep_params.hashAlg);
+                free(s);
+                continue;
+            }
+        }
+
         free(s);
 
         // clear buffers
@@ -191,11 +206,11 @@ CK_RV do_EncryptDecryptRSA(struct GENERATED_TEST_SUITE_INFO *tsuite)
         original_len = tsuite->tv[i].inputlen;
 
         // generate key pair
-        rc = generate_RSA_PKCS_KeyPair(session,
-                                       tsuite->tv[i].modbits,
-                                       tsuite->tv[i].publ_exp,
-                                       tsuite->tv[i].publ_exp_len,
-                                       &publ_key, &priv_key);
+        rc = generate_RSA_PKCS_KeyPair_cached(session,
+                                              tsuite->tv[i].modbits,
+                                              tsuite->tv[i].publ_exp,
+                                              tsuite->tv[i].publ_exp_len,
+                                              &publ_key, &priv_key);
 
         if (rc != CKR_OK) {
             if (rc == CKR_POLICY_VIOLATION) {
@@ -203,7 +218,7 @@ CK_RV do_EncryptDecryptRSA(struct GENERATED_TEST_SUITE_INFO *tsuite)
                 continue;
             }
 
-            testcase_error("generate_RSA_PKCS_KeyPair(), "
+            testcase_error("generate_RSA_PKCS_KeyPair_cached(), "
                            "rc=%s", p11_get_ckr(rc));
             goto testcase_cleanup;
         }
@@ -233,30 +248,56 @@ CK_RV do_EncryptDecryptRSA(struct GENERATED_TEST_SUITE_INFO *tsuite)
                  oaep_params.mgf != CKG_MGF1_SHA1)) {
                 testcase_skip("EP11 Token does not support RSA OAEP with hash "
                               "and/or MGF other than SHA-1");
-                goto tv_cleanup;
+                continue;
+            }
+
+            if (rc == CKR_MECHANISM_PARAM_INVALID &&
+                mech.mechanism == CKM_RSA_PKCS_OAEP &&
+                is_cca_token(slot_id) &&
+                (oaep_params.hashAlg != CKM_SHA_1 ||
+                 oaep_params.hashAlg != CKM_SHA256 ||
+                 oaep_params.mgf != CKG_MGF1_SHA1 ||
+                 oaep_params.mgf != CKG_MGF1_SHA256)) {
+                testcase_skip("CCA Token does only support RSA OAEP with hash "
+                              "and/or MGF other than SHA-1/SHA256 with "
+                              "CCA 8.1 or later");
+                continue;
             }
 
             testcase_error("C_EncryptInit, rc=%s", p11_get_ckr(rc));
-            goto error;
+            goto testcase_cleanup;
         }
         // do (public key) encryption
         rc = funcs->C_Encrypt(session,
                               original, original_len, crypt, &crypt_len);
         if (rc != CKR_OK) {
+            if (rc == CKR_MECHANISM_PARAM_INVALID &&
+                mech.mechanism == CKM_RSA_PKCS_OAEP &&
+                is_cca_token(slot_id) &&
+                (oaep_params.hashAlg != CKM_SHA_1 ||
+                 oaep_params.hashAlg != CKM_SHA256 ||
+                 oaep_params.mgf != CKG_MGF1_SHA1 ||
+                 oaep_params.mgf != CKG_MGF1_SHA256)) {
+                testcase_skip("CCA Token does only support RSA OAEP with hash "
+                              "and/or MGF other than SHA-1/SHA256 with "
+                              "CCA 8.1 or later");
+                continue;
+            }
+
             testcase_error("C_Encrypt, rc=%s", p11_get_ckr(rc));
-            goto error;
+            goto testcase_cleanup;
         }
         // initialize (private key) decryption
         rc = funcs->C_DecryptInit(session, &mech, priv_key);
         if (rc != CKR_OK) {
             testcase_error("C_DecryptInit, rc=%s", p11_get_ckr(rc));
-            goto error;
+            goto testcase_cleanup;
         }
         // do (private key) decryption
         rc = funcs->C_Decrypt(session, crypt, crypt_len, decrypt, &decrypt_len);
         if (rc != CKR_OK) {
             testcase_error("C_Decrypt, rc=%s", p11_get_ckr(rc));
-            goto error;
+            goto testcase_cleanup;
         }
         // FIXME: there shouldn't be any padding here
         // remove padding if mech is CKM_RSA_X_509
@@ -278,34 +319,12 @@ CK_RV do_EncryptDecryptRSA(struct GENERATED_TEST_SUITE_INFO *tsuite)
             testcase_pass("C_Encrypt and C_Decrypt.");
         }
 
-        // clean up
-tv_cleanup:
-        rc = funcs->C_DestroyObject(session, publ_key);
-        if (rc != CKR_OK) {
-            testcase_error("C_DestroyObject(), rc=%s.", p11_get_ckr(rc));
-            goto error;
-        }
-
-        rc = funcs->C_DestroyObject(session, priv_key);
-        if (rc != CKR_OK) {
-            testcase_error("C_DestroyObject(), rc=%s.", p11_get_ckr(rc));
-            goto error;
-        }
     }
 
     goto testcase_cleanup;
-error:
-    loc_rc = funcs->C_DestroyObject(session, publ_key);
-    if (loc_rc != CKR_OK) {
-        testcase_error("C_DestroyObject(), rc=%s.", p11_get_ckr(loc_rc));
-    }
-
-    loc_rc = funcs->C_DestroyObject(session, priv_key);
-    if (loc_rc != CKR_OK) {
-        testcase_error("C_DestroyObject(), rc=%s.", p11_get_ckr(loc_rc));
-    }
 
 testcase_cleanup:
+    free_rsa_key_cache(session);
     testcase_user_logout();
     loc_rc = funcs->C_CloseAllSessions(slot_id);
     if (loc_rc != CKR_OK) {
@@ -333,7 +352,8 @@ testcase_cleanup:
  * 5. Compare plaintext with decrypted data
  *
  */
-CK_RV do_EncryptDecryptImportRSA(struct PUBLISHED_TEST_SUITE_INFO *tsuite)
+CK_RV do_EncryptDecryptImportRSA(struct PUBLISHED_TEST_SUITE_INFO *tsuite,
+                                 CK_BBOOL me_only)
 {
     unsigned int i, j;
     CK_BYTE original[BIG_REQUEST];
@@ -354,7 +374,8 @@ CK_RV do_EncryptDecryptImportRSA(struct PUBLISHED_TEST_SUITE_INFO *tsuite)
     char *s;
 
     // begin testsuite
-    testsuite_begin("%s Encrypt Decrypt Import.", tsuite->name);
+    testsuite_begin("%s Encrypt Decrypt Import%s.", tsuite->name,
+                    me_only ? " (ME-format)" : "");
     testcase_rw_session();
     testcase_user_login();
 
@@ -367,6 +388,35 @@ CK_RV do_EncryptDecryptImportRSA(struct PUBLISHED_TEST_SUITE_INFO *tsuite)
                        (unsigned int) tsuite->mech.mechanism);
         goto testcase_cleanup;
     }
+
+    if (is_cca_token(slot_id) && tsuite->mech.mechanism == CKM_RSA_PKCS_OAEP) {
+        if (((CK_RSA_PKCS_OAEP_PARAMS *)tsuite->mech.pParameter)->hashAlg !=
+                                                                CKM_SHA_1 &&
+            ((CK_RSA_PKCS_OAEP_PARAMS *)tsuite->mech.pParameter)->hashAlg !=
+                                                                CKM_SHA224 &&
+            ((CK_RSA_PKCS_OAEP_PARAMS *)tsuite->mech.pParameter)->hashAlg !=
+                                                                CKM_SHA256 &&
+            ((CK_RSA_PKCS_OAEP_PARAMS *)tsuite->mech.pParameter)->hashAlg !=
+                                                                CKM_SHA384 &&
+            ((CK_RSA_PKCS_OAEP_PARAMS *)tsuite->mech.pParameter)->hashAlg !=
+                                                                CKM_SHA512) {
+             testcase_skip("CCA Token cannot use RSA OAEP with a hash "
+                          "algorithm other than SHA1 and SHA2: %s",
+                          mech_to_str(((CK_RSA_PKCS_OAEP_PARAMS *)
+                                          tsuite->mech.pParameter)->hashAlg));
+             goto testcase_cleanup;
+         }
+
+        if (((CK_RSA_PKCS_OAEP_PARAMS *)tsuite->mech.pParameter)->source ==
+                                                        CKZ_DATA_SPECIFIED &&
+            ((CK_RSA_PKCS_OAEP_PARAMS *)tsuite->mech.pParameter)->ulSourceDataLen
+                                                        > 0) {
+             testcase_skip("CCA Token cannot use RSA OAEP with non empty "
+                           "source data");
+             goto testcase_cleanup;
+         }
+    }
+
     // iterate over test vectors
     for (i = 0; i < tsuite->tvcount; i++) {
 
@@ -378,9 +428,9 @@ CK_RV do_EncryptDecryptImportRSA(struct PUBLISHED_TEST_SUITE_INFO *tsuite)
             goto testcase_cleanup;
         }
         // begin testcase
-        testcase_begin("%s Encrypt and Decrypt Import with test vector %u."
+        testcase_begin("%s Encrypt and Decrypt Import with test vector %u%s."
                        "\npubl_exp='%s', modbits=%lu, publ_exp_len=%lu.",
-                       tsuite->name, i, s,
+                       tsuite->name, i, me_only ? " (ME-format)" : "", s,
                        tsuite->tv[i].mod_len * 8,
                        tsuite->tv[i].pubexp_len);
 
@@ -394,6 +444,16 @@ CK_RV do_EncryptDecryptImportRSA(struct PUBLISHED_TEST_SUITE_INFO *tsuite)
             continue;
         }
 
+        if (tsuite->mech.mechanism == CKM_RSA_PKCS_OAEP &&
+            ((CK_RSA_PKCS_OAEP_PARAMS *)tsuite->mech.pParameter)->hashAlg ==
+                                                                   CKM_SHA512 &&
+            tsuite->tv[i].mod_len * 8 <= 1024) {
+            testcase_skip("OAEP with SHA512 cannot be used with modbits='%lu'",
+                          tsuite->tv[i].mod_len * 8);
+            free(s);
+            continue;
+        }
+
         if (is_ep11_token(slot_id) || is_icsf_token(slot_id)) {
             if (!is_valid_ep11_pubexp(tsuite->tv[i].pub_exp,
                                       tsuite->tv[i].pubexp_len)) {
@@ -492,19 +552,20 @@ CK_RV do_EncryptDecryptImportRSA(struct PUBLISHED_TEST_SUITE_INFO *tsuite)
                                   tsuite->tv[i].mod,
                                   tsuite->tv[i].pub_exp,
                                   tsuite->tv[i].priv_exp,
-                                  tsuite->tv[i].prime1,
-                                  tsuite->tv[i].prime2,
-                                  tsuite->tv[i].exp1,
-                                  tsuite->tv[i].exp2,
-                                  tsuite->tv[i].coef,
+                                  me_only ? NULL : tsuite->tv[i].prime1,
+                                  me_only ? NULL : tsuite->tv[i].prime2,
+                                  me_only ? NULL : tsuite->tv[i].exp1,
+                                  me_only ? NULL : tsuite->tv[i].exp2,
+                                  me_only ? NULL : tsuite->tv[i].coef,
                                   tsuite->tv[i].mod_len,
                                   tsuite->tv[i].pubexp_len,
                                   tsuite->tv[i].privexp_len,
-                                  tsuite->tv[i].prime1_len,
-                                  tsuite->tv[i].prime2_len,
-                                  tsuite->tv[i].exp1_len,
-                                  tsuite->tv[i].exp2_len,
-                                  tsuite->tv[i].coef_len, &priv_key);
+                                  me_only ? 0 : tsuite->tv[i].prime1_len,
+                                  me_only ? 0 : tsuite->tv[i].prime2_len,
+                                  me_only ? 0 : tsuite->tv[i].exp1_len,
+                                  me_only ? 0 : tsuite->tv[i].exp2_len,
+                                  me_only ? 0 : tsuite->tv[i].coef_len,
+                                  &priv_key);
         if (rc != CKR_OK) {
             if (rc == CKR_POLICY_VIOLATION) {
                 testcase_skip("RSA key import is not allowed by policy");
@@ -547,16 +608,33 @@ CK_RV do_EncryptDecryptImportRSA(struct PUBLISHED_TEST_SUITE_INFO *tsuite)
         rc = funcs->C_EncryptInit(session, &mech, publ_key);
         if (rc != CKR_OK) {
             if (rc == CKR_MECHANISM_PARAM_INVALID &&
-                 mech.mechanism == CKM_RSA_PKCS_OAEP &&
-                 is_ep11_token(slot_id) &&
-                 (((CK_RSA_PKCS_OAEP_PARAMS *)mech.pParameter)->hashAlg !=
-                                                              CKM_SHA_1 ||
-                  ((CK_RSA_PKCS_OAEP_PARAMS *)mech.pParameter)->mgf !=
-                                                              CKG_MGF1_SHA1)) {
-                 testcase_skip("EP11 Token does not support RSA OAEP with hash "
-                               "and/or MGF other than SHA-1");
-                 goto tv_cleanup;
-             }
+                mech.mechanism == CKM_RSA_PKCS_OAEP &&
+                is_ep11_token(slot_id) &&
+                (((CK_RSA_PKCS_OAEP_PARAMS *)mech.pParameter)->hashAlg !=
+                                                             CKM_SHA_1 ||
+                 ((CK_RSA_PKCS_OAEP_PARAMS *)mech.pParameter)->mgf !=
+                                                             CKG_MGF1_SHA1)) {
+                testcase_skip("EP11 Token does not support RSA OAEP with hash "
+                              "and/or MGF other than SHA-1");
+                goto tv_cleanup;
+            }
+
+            if (rc == CKR_MECHANISM_PARAM_INVALID &&
+                mech.mechanism == CKM_RSA_PKCS_OAEP &&
+                is_cca_token(slot_id) &&
+                (((CK_RSA_PKCS_OAEP_PARAMS *)mech.pParameter)->hashAlg !=
+                                                                CKM_SHA_1 ||
+                 ((CK_RSA_PKCS_OAEP_PARAMS *)mech.pParameter)->hashAlg !=
+                                                                 CKM_SHA256 ||
+                 ((CK_RSA_PKCS_OAEP_PARAMS *)mech.pParameter)->mgf !=
+                                                             CKG_MGF1_SHA1 ||
+                 ((CK_RSA_PKCS_OAEP_PARAMS *)mech.pParameter)->mgf !=
+                                                             CKG_MGF1_SHA256)) {
+                testcase_skip("CCA Token does only support RSA OAEP with hash "
+                              "and/or MGF other than SHA-1/SHA256 with "
+                              "CCA 8.1 or later");
+                goto tv_cleanup;
+            }
 
             testcase_error("C_EncryptInit, rc=%s", p11_get_ckr(rc));
             goto tv_cleanup;
@@ -565,6 +643,23 @@ CK_RV do_EncryptDecryptImportRSA(struct PUBLISHED_TEST_SUITE_INFO *tsuite)
         rc = funcs->C_Encrypt(session,
                               original, original_len, crypt, &crypt_len);
         if (rc != CKR_OK) {
+            if (rc == CKR_MECHANISM_PARAM_INVALID &&
+                mech.mechanism == CKM_RSA_PKCS_OAEP &&
+                is_cca_token(slot_id) &&
+                (((CK_RSA_PKCS_OAEP_PARAMS *)mech.pParameter)->hashAlg !=
+                                                                CKM_SHA_1 ||
+                 ((CK_RSA_PKCS_OAEP_PARAMS *)mech.pParameter)->hashAlg !=
+                                                                 CKM_SHA256 ||
+                 ((CK_RSA_PKCS_OAEP_PARAMS *)mech.pParameter)->mgf !=
+                                                             CKG_MGF1_SHA1 ||
+                 ((CK_RSA_PKCS_OAEP_PARAMS *)mech.pParameter)->mgf !=
+                                                             CKG_MGF1_SHA256)) {
+                testcase_skip("CCA Token does only support RSA OAEP with hash "
+                              "and/or MGF other than SHA-1/SHA256 with "
+                              "CCA 8.1 or later");
+                goto tv_cleanup;
+            }
+
             testcase_error("C_Encrypt, rc=%s", p11_get_ckr(rc));
             goto tv_cleanup;
         }
@@ -597,7 +692,8 @@ CK_RV do_EncryptDecryptImportRSA(struct PUBLISHED_TEST_SUITE_INFO *tsuite)
         } else if (memcmp(decrypt, original, original_len)) {
             testcase_fail("decrypted data does not match " "original data.");
         } else {
-            testcase_pass("C_Encrypt and C_Decrypt.");
+            testcase_pass("C_Encrypt and C_Decrypt%s.",
+                          me_only ? " (ME-format)" : "");
         }
 
         // clean up
@@ -668,7 +764,7 @@ CK_RV do_SignVerifyRSA(struct GENERATED_TEST_SUITE_INFO * tsuite,
     CK_FLAGS flags;
     CK_BYTE user_pin[PKCS11_MAX_PIN_LEN];
     CK_ULONG user_pin_len;
-    CK_RV rc, loc_rc;
+    CK_RV rc;
 
     char *s;
 
@@ -770,6 +866,19 @@ CK_RV do_SignVerifyRSA(struct GENERATED_TEST_SUITE_INFO * tsuite,
                 continue;
             }
         }
+
+        if (tsuite->tv[i].modbits <= 512 &&
+            (tsuite->mech.mechanism == CKM_SHA384_RSA_PKCS ||
+             tsuite->mech.mechanism == CKM_SHA512_RSA_PKCS ||
+             tsuite->mech.mechanism == CKM_SHA3_384_RSA_PKCS ||
+             tsuite->mech.mechanism == CKM_SHA3_512_RSA_PKCS)) {
+            testcase_skip("Mechanism %s can not be used with a key with mod_bits='%lu'.",
+                          mech_to_str(tsuite->mech.mechanism),
+                          tsuite->tv[i].modbits);
+            free(s);
+            continue;
+        }
+
         // free memory
         free(s);
 
@@ -784,18 +893,18 @@ CK_RV do_SignVerifyRSA(struct GENERATED_TEST_SUITE_INFO * tsuite,
         message_len = tsuite->tv[i].inputlen;
 
         // generate key pair
-        rc = generate_RSA_PKCS_KeyPair(session,
-                                       tsuite->tv[i].modbits,
-                                       tsuite->tv[i].publ_exp,
-                                       tsuite->tv[i].publ_exp_len,
-                                       &publ_key, &priv_key);
+        rc = generate_RSA_PKCS_KeyPair_cached(session,
+                                              tsuite->tv[i].modbits,
+                                              tsuite->tv[i].publ_exp,
+                                              tsuite->tv[i].publ_exp_len,
+                                              &publ_key, &priv_key);
         if (rc != CKR_OK) {
             if (rc == CKR_POLICY_VIOLATION) {
                 testcase_skip("RSA key generation is not allowed by policy");
                 continue;
             }
 
-            testcase_error("generate_RSA_PKCS_KeyPair(), "
+            testcase_error("generate_RSA_PKCS_KeyPair_cached(), "
                            "rc=%s", p11_get_ckr(rc));
             goto testcase_cleanup;
         }
@@ -815,7 +924,7 @@ CK_RV do_SignVerifyRSA(struct GENERATED_TEST_SUITE_INFO * tsuite,
         if (rc != CKR_OK) {
             testcase_error("C_Sign%sInit(), rc=%s",
                            recover_mode ? "Recover" : "", p11_get_ckr(rc));
-            goto error;
+            goto testcase_cleanup;
         }
         // set buffer size
         signature_len = MAX_SIGNATURE_SIZE;
@@ -831,7 +940,7 @@ CK_RV do_SignVerifyRSA(struct GENERATED_TEST_SUITE_INFO * tsuite,
             testcase_error("C_Sign%s(), rc=%s signature len=%lu",
                            recover_mode ? "Recover" : "",
                            p11_get_ckr(rc), signature_len);
-            goto error;
+            goto testcase_cleanup;
         }
         // initialize Verify
         if (recover_mode)
@@ -841,7 +950,7 @@ CK_RV do_SignVerifyRSA(struct GENERATED_TEST_SUITE_INFO * tsuite,
         if (rc != CKR_OK) {
             testcase_error("C_Verify%sInit(), rc=%s",
                            recover_mode ? "Recover" : "", p11_get_ckr(rc));
-            goto error;
+            goto testcase_cleanup;
         }
         // do Verify
         if (recover_mode) {
@@ -881,29 +990,10 @@ CK_RV do_SignVerifyRSA(struct GENERATED_TEST_SUITE_INFO * tsuite,
                           p11_get_ckr(rc));
         }
 
-        // clean up
-        rc = funcs->C_DestroyObject(session, publ_key);
-        if (rc != CKR_OK) {
-            testcase_error("C_DestroyObject(), rc=%s.", p11_get_ckr(rc));
-        }
-
-        rc = funcs->C_DestroyObject(session, priv_key);
-        if (rc != CKR_OK) {
-            testcase_error("C_DestroyObject(), rc=%s.", p11_get_ckr(rc));
-        }
-    }
-    goto testcase_cleanup;
-error:
-    loc_rc = funcs->C_DestroyObject(session, publ_key);
-    if (loc_rc != CKR_OK) {
-        testcase_error("C_DestroyObject, rc=%s.", p11_get_ckr(loc_rc));
-    }
-    loc_rc = funcs->C_DestroyObject(session, priv_key);
-    if (loc_rc != CKR_OK) {
-        testcase_error("C_DestroyObject, rc=%s.", p11_get_ckr(loc_rc));
     }
 
 testcase_cleanup:
+    free_rsa_key_cache(session);
     testcase_user_logout();
     rc = funcs->C_CloseAllSessions(slot_id);
     if (rc != CKR_OK) {
@@ -944,7 +1034,7 @@ CK_RV do_SignVerify_RSAPSS(struct GENERATED_TEST_SUITE_INFO * tsuite)
     CK_FLAGS flags;
     CK_BYTE user_pin[PKCS11_MAX_PIN_LEN];
     CK_ULONG user_pin_len;
-    CK_RV rc, loc_rc;
+    CK_RV rc;
     CK_RSA_PKCS_PSS_PARAMS pss_params;
 
     char *s;
@@ -1015,19 +1105,19 @@ CK_RV do_SignVerify_RSAPSS(struct GENERATED_TEST_SUITE_INFO * tsuite)
         message_len = tsuite->tv[i].inputlen;
 
         // generate key pair
-        rc = generate_RSA_PKCS_KeyPair(session, tsuite->tv[i].modbits,
-                                       tsuite->tv[i].publ_exp,
-                                       tsuite->tv[i].publ_exp_len, &publ_key,
-                                       &priv_key);
+        rc = generate_RSA_PKCS_KeyPair_cached(session, tsuite->tv[i].modbits,
+                                              tsuite->tv[i].publ_exp,
+                                              tsuite->tv[i].publ_exp_len,
+                                              &publ_key, &priv_key);
         if (rc != CKR_OK) {
             if (rc == CKR_POLICY_VIOLATION) {
                 testcase_skip("RSA key generation is not allowed by policy");
                 continue;
             }
 
-            testcase_error("generate_RSA_PKCS_KeyPair(), "
+            testcase_error("generate_RSA_PKCS_KeyPair_cached(), "
                            "rc=%s", p11_get_ckr(rc));
-            goto error;
+            goto testcase_cleanup;
         }
         // generate message
         for (j = 0; j < message_len; j++) {
@@ -1047,7 +1137,7 @@ CK_RV do_SignVerify_RSAPSS(struct GENERATED_TEST_SUITE_INFO * tsuite)
                               (unsigned int)slot_id,
                               mech_to_str(mech.mechanism),
                               (unsigned int)mech.mechanism);
-                goto cleanup;
+                continue;
             }
 
             rc = funcs->C_DigestInit(session, &mech);
@@ -1071,7 +1161,7 @@ CK_RV do_SignVerify_RSAPSS(struct GENERATED_TEST_SUITE_INFO * tsuite)
         rc = funcs->C_SignInit(session, &mech, priv_key);
         if (rc != CKR_OK) {
             testcase_error("C_SignInit(), rc=%s", p11_get_ckr(rc));
-            goto error;
+            goto testcase_cleanup;
         }
         // set buffer size
         signature_len = MAX_SIGNATURE_SIZE;
@@ -1085,7 +1175,7 @@ CK_RV do_SignVerify_RSAPSS(struct GENERATED_TEST_SUITE_INFO * tsuite)
         if (rc != CKR_OK) {
             testcase_error("C_Sign(), rc=%s signature len=%lu",
                            p11_get_ckr(rc), signature_len);
-            goto error;
+            goto testcase_cleanup;
         }
         // initialize Verify
         rc = funcs->C_VerifyInit(session, &mech, publ_key);
@@ -1106,31 +1196,10 @@ CK_RV do_SignVerify_RSAPSS(struct GENERATED_TEST_SUITE_INFO * tsuite)
             testcase_pass("C_Verify.");
         else
             testcase_fail("C_Verify(), rc=%s", p11_get_ckr(rc));
-
-cleanup:
-        // clean up
-        rc = funcs->C_DestroyObject(session, publ_key);
-        if (rc != CKR_OK) {
-            testcase_error("C_DestroyObject(), rc=%s.", p11_get_ckr(rc));
-        }
-
-        rc = funcs->C_DestroyObject(session, priv_key);
-        if (rc != CKR_OK) {
-            testcase_error("C_DestroyObject(), rc=%s.", p11_get_ckr(rc));
-        }
-    }
-    goto testcase_cleanup;
-error:
-    loc_rc = funcs->C_DestroyObject(session, publ_key);
-    if (loc_rc != CKR_OK) {
-        testcase_error("C_DestroyObject, rc=%s.", p11_get_ckr(loc_rc));
-    }
-    loc_rc = funcs->C_DestroyObject(session, priv_key);
-    if (loc_rc != CKR_OK) {
-        testcase_error("C_DestroyObject, rc=%s.", p11_get_ckr(loc_rc));
     }
 
 testcase_cleanup:
+    free_rsa_key_cache(session);
     testcase_user_logout();
     rc = funcs->C_CloseAllSessions(slot_id);
     if (rc != CKR_OK) {
@@ -1268,7 +1337,8 @@ CK_RV do_WrapUnwrapRSA(struct GENERATED_TEST_SUITE_INFO * tsuite)
                 tsuite->tv[i].oaep_params.hashAlg != CKM_SHA_1 &&
                 tsuite->tv[i].oaep_params.hashAlg != CKM_SHA256) {
                 testcase_skip("CCA Token cannot use RSA OAEP with a hash "
-                             "algorithm other than SHA1 and SHA256");
+                             "algorithm other than SHA1 and SHA256: %s",
+                             mech_to_str(tsuite->tv[i].oaep_params.hashAlg));
                 free(s);
                 continue;
             }
@@ -1291,6 +1361,17 @@ CK_RV do_WrapUnwrapRSA(struct GENERATED_TEST_SUITE_INFO * tsuite)
             }
         }
 
+        if (tsuite->mech.mechanism == CKM_RSA_PKCS_OAEP) {
+            if (!mech_supported(slot_id, tsuite->tv[i].oaep_params.hashAlg)) {
+                testcase_skip("Slot %u doesn't support OAEP hash alg %s (0x%x)",
+                              (unsigned int)slot_id,
+                              mech_to_str(tsuite->tv[i].oaep_params.hashAlg),
+                              (unsigned int)tsuite->tv[i].oaep_params.hashAlg);
+                free(s);
+                continue;
+            }
+        }
+
         // begin test
         testcase_begin("%s Wrap Unwrap with test vector %u, "
                        "\npubl_exp='%s', mod_bits='%lu', keylen='%lu', "
@@ -1343,10 +1424,10 @@ CK_RV do_WrapUnwrapRSA(struct GENERATED_TEST_SUITE_INFO * tsuite)
         wrapped_keylen = PKCS11_MAX_PIN_LEN;
 
         // generate RSA key pair
-        rc = generate_RSA_PKCS_KeyPair(session, tsuite->tv[i].modbits,
-                                       tsuite->tv[i].publ_exp,
-                                       tsuite->tv[i].publ_exp_len,
-                                       &publ_key, &priv_key);
+        rc = generate_RSA_PKCS_KeyPair_cached(session, tsuite->tv[i].modbits,
+                                              tsuite->tv[i].publ_exp,
+                                              tsuite->tv[i].publ_exp_len,
+                                              &publ_key, &priv_key);
         if (rc != CKR_OK) {
             if (rc == CKR_POLICY_VIOLATION) {
                 testcase_skip("RSA key generation is not allowed by policy");
@@ -1463,7 +1544,7 @@ CK_RV do_WrapUnwrapRSA(struct GENERATED_TEST_SUITE_INFO * tsuite)
         testcase_new_assertion();       /* assertion #1 */
 
         // allocate memory for wrapped_key
-        wrapped_key = calloc(sizeof(CK_BYTE), wrapped_keylen);
+        wrapped_key = calloc(wrapped_keylen, sizeof(CK_BYTE));
         if (wrapped_key == NULL) {
             testcase_error("Can't allocate memory for %lu bytes.",
                            sizeof(CK_BYTE) * wrapped_keylen);
@@ -1558,13 +1639,6 @@ tv_cleanup:
         if (rc != CKR_OK)
             testcase_error("C_DestroyObject(), rc=%s.", p11_get_ckr(rc));
 
-        rc = funcs->C_DestroyObject(session, publ_key);
-        if (rc != CKR_OK)
-            testcase_error("C_DestroyObject(), rc=%s.", p11_get_ckr(rc));
-
-        rc = funcs->C_DestroyObject(session, priv_key);
-        if (rc != CKR_OK)
-            testcase_error("C_DestroyObject(), rc=%s.", p11_get_ckr(rc));
     }
     goto testcase_cleanup;
 
@@ -1575,10 +1649,9 @@ error:
     }
 
     funcs->C_DestroyObject(session, secret_key);
-    funcs->C_DestroyObject(session, publ_key);
-    funcs->C_DestroyObject(session, priv_key);
 
 testcase_cleanup:
+    free_rsa_key_cache(session);
     testcase_user_logout();
     loc_rc = funcs->C_CloseAllSessions(slot_id);
     if (loc_rc != CKR_OK) {
@@ -1598,7 +1671,7 @@ testcase_cleanup:
  * 4. Compare expected signature with actual signature
  *
  */
-CK_RV do_SignRSA(struct PUBLISHED_TEST_SUITE_INFO * tsuite)
+CK_RV do_SignRSA(struct PUBLISHED_TEST_SUITE_INFO * tsuite, CK_BBOOL me_only)
 {
     unsigned int i;
     CK_BYTE message[MAX_MESSAGE_SIZE];
@@ -1617,7 +1690,7 @@ CK_RV do_SignRSA(struct PUBLISHED_TEST_SUITE_INFO * tsuite)
     CK_RV rc, loc_rc;
 
     // begin testsuite
-    testsuite_begin("%s Sign. ", tsuite->name);
+    testsuite_begin("%s Sign%s. ", tsuite->name, me_only ? " (ME-format)" : "");
     testcase_rw_session();
     testcase_user_login();
 
@@ -1632,7 +1705,8 @@ CK_RV do_SignRSA(struct PUBLISHED_TEST_SUITE_INFO * tsuite)
     }
     // iterate over test vectors
     for (i = 0; i < tsuite->tvcount; i++) {
-        testcase_begin("%s Sign with test vector %u.", tsuite->name, i);
+        testcase_begin("%s Sign%s with test vector %u.", tsuite->name,
+                       me_only ? " (ME-format)" : "", i);
 
         rc = CKR_OK;            // set return value
 
@@ -1719,19 +1793,20 @@ CK_RV do_SignRSA(struct PUBLISHED_TEST_SUITE_INFO * tsuite)
                                   tsuite->tv[i].mod,
                                   tsuite->tv[i].pub_exp,
                                   tsuite->tv[i].priv_exp,
-                                  tsuite->tv[i].prime1,
-                                  tsuite->tv[i].prime2,
-                                  tsuite->tv[i].exp1,
-                                  tsuite->tv[i].exp2,
-                                  tsuite->tv[i].coef,
+                                  me_only ? NULL : tsuite->tv[i].prime1,
+                                  me_only ? NULL : tsuite->tv[i].prime2,
+                                  me_only ? NULL : tsuite->tv[i].exp1,
+                                  me_only ? NULL : tsuite->tv[i].exp2,
+                                  me_only ? NULL : tsuite->tv[i].coef,
                                   tsuite->tv[i].mod_len,
                                   tsuite->tv[i].pubexp_len,
                                   tsuite->tv[i].privexp_len,
-                                  tsuite->tv[i].prime1_len,
-                                  tsuite->tv[i].prime2_len,
-                                  tsuite->tv[i].exp1_len,
-                                  tsuite->tv[i].exp2_len,
-                                  tsuite->tv[i].coef_len, &priv_key);
+                                  me_only ? 0 : tsuite->tv[i].prime1_len,
+                                  me_only ? 0 : tsuite->tv[i].prime2_len,
+                                  me_only ? 0 : tsuite->tv[i].exp1_len,
+                                  me_only ? 0 : tsuite->tv[i].exp2_len,
+                                  me_only ? 0 : tsuite->tv[i].coef_len,
+                                  &priv_key);
         if (rc != CKR_OK) {
             if (rc == CKR_POLICY_VIOLATION) {
                 testcase_skip("RSA key import is not allowed by policy");
@@ -1770,7 +1845,7 @@ CK_RV do_SignRSA(struct PUBLISHED_TEST_SUITE_INFO * tsuite)
                           "signature.", tsuite->name, i);
 
         } else {
-            testcase_pass("C_Sign.");
+            testcase_pass("C_Sign%s.", me_only ? " (ME-format)" : "");
         }
 skip:
         // clean up
@@ -1963,6 +2038,164 @@ testcase_cleanup:
     return rc;
 }
 
+CK_RV do_RSAImplicitRejection(struct PUBLISHED_TEST_SUITE_INFO *tsuite)
+{
+    unsigned int i;
+    CK_BYTE decrypt[BIG_REQUEST];
+    CK_ULONG decrypt_len;
+    CK_MECHANISM mech;
+    CK_OBJECT_HANDLE priv_key = CK_INVALID_HANDLE;
+    CK_SLOT_ID slot_id = SLOT_ID;
+    CK_SESSION_HANDLE session;
+    CK_FLAGS flags;
+    CK_BYTE user_pin[PKCS11_MAX_PIN_LEN];
+    CK_ULONG user_pin_len;
+    CK_RV rc, loc_rc;
+
+    char *s;
+
+    // begin testsuite
+    testsuite_begin("%s Implicit Rejection.", tsuite->name);
+    testcase_rw_session();
+    testcase_user_login();
+
+    if (!is_ica_token(slot_id) && !is_soft_token(slot_id)) {
+        testsuite_skip(tsuite->tvcount,
+                       "Slot %u doesn't support Implicit Rejection",
+                       (unsigned int) slot_id);
+        goto testcase_cleanup;
+    }
+    // skip tests if the slot doesn't support this mechanism
+    if (!mech_supported(slot_id, tsuite->mech.mechanism)) {
+        testsuite_skip(tsuite->tvcount,
+                       "Slot %u doesn't support %s (0x%x)",
+                       (unsigned int) slot_id,
+                       mech_to_str(tsuite->mech.mechanism),
+                       (unsigned int) tsuite->mech.mechanism);
+        goto testcase_cleanup;
+    }
+
+    // iterate over test vectors
+    for (i = 0; i < tsuite->tvcount; i++) {
+
+        // get public exponent from test vector
+        if (p11_ahex_dump(&s, tsuite->tv[i].pub_exp,
+                          tsuite->tv[i].pubexp_len) == NULL) {
+            testcase_error("p11_ahex_dump() failed");
+            rc = -1;
+            goto testcase_cleanup;
+        }
+        // begin testcase
+        testcase_begin("%s Implicit Rejection with test vector %u."
+                       "\npubl_exp='%s', modbits=%lu, publ_exp_len=%lu.",
+                       tsuite->name, i, s,
+                       tsuite->tv[i].mod_len * 8,
+                       tsuite->tv[i].pubexp_len);
+
+        rc = CKR_OK;
+
+        if (!keysize_supported(slot_id, tsuite->mech.mechanism,
+                               tsuite->tv[i].mod_len * 8)) {
+            testcase_skip("Token in slot %lu cannot be used with modbits='%lu'",
+                          SLOT_ID, tsuite->tv[i].mod_len * 8);
+            free(s);
+            continue;
+        }
+
+        free(s);
+
+        // clear buffers
+        memset(decrypt, 0, BIG_REQUEST);
+
+        // create (private) key handle
+        rc = create_RSAPrivateKey(session,
+                                  tsuite->tv[i].mod,
+                                  tsuite->tv[i].pub_exp,
+                                  tsuite->tv[i].priv_exp,
+                                  tsuite->tv[i].prime1,
+                                  tsuite->tv[i].prime2,
+                                  tsuite->tv[i].exp1,
+                                  tsuite->tv[i].exp2,
+                                  tsuite->tv[i].coef,
+                                  tsuite->tv[i].mod_len,
+                                  tsuite->tv[i].pubexp_len,
+                                  tsuite->tv[i].privexp_len,
+                                  tsuite->tv[i].prime1_len,
+                                  tsuite->tv[i].prime2_len,
+                                  tsuite->tv[i].exp1_len,
+                                  tsuite->tv[i].exp2_len,
+                                  tsuite->tv[i].coef_len, &priv_key);
+        if (rc != CKR_OK) {
+            if (rc == CKR_POLICY_VIOLATION) {
+                testcase_skip("RSA key import is not allowed by policy");
+                continue;
+            }
+
+            testcase_error("create_RSAPrivateKey(), rc=%s", p11_get_ckr(rc));
+            goto error;
+        }
+
+        // set cipher buffer length
+        decrypt_len = BIG_REQUEST;
+
+        // get mech
+        mech = tsuite->mech;
+
+        // initialize (private key) decryption
+        rc = funcs->C_DecryptInit(session, &mech, priv_key);
+        if (rc != CKR_OK) {
+            testcase_error("C_DecryptInit, rc=%s", p11_get_ckr(rc));
+            goto tv_cleanup;
+        }
+        // do (private key) decryption
+        rc = funcs->C_Decrypt(session, tsuite->tv[i].msg, tsuite->tv[i].msg_len,
+                              decrypt, &decrypt_len);
+        if (rc != CKR_OK) {
+            testcase_error("C_Decrypt, rc=%s", p11_get_ckr(rc));
+            goto tv_cleanup;
+        }
+
+        // check results
+        testcase_new_assertion();
+
+        if (decrypt_len != tsuite->tv[i].sig_len) {
+            testcase_fail("decrypted length does not match"
+                          "expected data length.\n expected length = %lu, "
+                          "but found length=%lu.\n",
+                          tsuite->tv[i].sig_len, decrypt_len);
+        } else if (memcmp(decrypt, tsuite->tv[i].sig, tsuite->tv[i].sig_len)) {
+            testcase_fail("decrypted data does not match expected data.");
+        } else {
+            testcase_pass("Implicit Rejection.");
+        }
+
+        // clean up
+tv_cleanup:
+
+        rc = funcs->C_DestroyObject(session, priv_key);
+        if (rc != CKR_OK) {
+            testcase_error("C_DestroyObject(), rc=%s.", p11_get_ckr(rc));
+            goto error;
+        }
+    }
+
+    goto testcase_cleanup;
+error:
+    loc_rc = funcs->C_DestroyObject(session, priv_key);
+    if (loc_rc != CKR_OK) {
+        testcase_error("C_DestroyObject(), rc=%s.", p11_get_ckr(loc_rc));
+    }
+
+testcase_cleanup:
+    testcase_user_logout();
+    loc_rc = funcs->C_CloseAllSessions(slot_id);
+    if (loc_rc != CKR_OK) {
+        testcase_error("C_CloseAllSessions, rc=%s", p11_get_ckr(loc_rc));
+    }
+
+    return rc;
+}
+
 CK_RV rsa_funcs(void)
 {
     unsigned int i;
@@ -1970,7 +2203,11 @@ CK_RV rsa_funcs(void)
 
     // published (known answer) tests
     for (i = 0; i < NUM_OF_PUBLISHED_TESTSUITES; i++) {
-        rv = do_SignRSA(&published_test_suites[i]);
+        rv = do_SignRSA(&published_test_suites[i], FALSE);
+        if (rv != CKR_OK && (!no_stop))
+            break;
+
+        rv = do_SignRSA(&published_test_suites[i], TRUE);
         if (rv != CKR_OK && (!no_stop))
             break;
 
@@ -2027,7 +2264,22 @@ CK_RV rsa_funcs(void)
 
     // key import tests
     for (i = 0; i < NUM_OF_ENCDEC_IMPORT_TESTSUITES; i++) {
-        rv = do_EncryptDecryptImportRSA(&rsa_encdec_import_test_suites[i]);
+        rv = do_EncryptDecryptImportRSA(&rsa_encdec_import_test_suites[i],
+                                        FALSE);
+        if (rv != CKR_OK && (!no_stop))
+            break;
+    }
+
+    for (i = 0; i < NUM_OF_ENCDEC_IMPORT_TESTSUITES; i++) {
+        rv = do_EncryptDecryptImportRSA(&rsa_encdec_import_test_suites[i],
+                                        TRUE);
+        if (rv != CKR_OK && (!no_stop))
+            break;
+    }
+
+    // Implicit rejection tests
+    for (i = 0; i < NUM_OF_IMPLICIT_REJECTION_TESTSUITES; i++) {
+        rv = do_RSAImplicitRejection(&rsa_implicit_rejection_test_suites[i]);
         if (rv != CKR_OK && (!no_stop))
             break;
     }
@@ -2076,6 +2328,7 @@ int main(int argc, char **argv)
     rv = rsa_funcs();
     testcase_print_result();
 
+    free_rsa_key_cache(CK_INVALID_HANDLE);
     funcs->C_Finalize(NULL);
 
     return testcase_return(rv);
